import pytest

from tradex.constants.enuming import ProfitPossibility
from tradex.models.order import OrderProfitFormSimple
from tradex.models.take_profit import TakeProfit, TakeProfitFibPer
from tradex.utils.profit import get_take_profit_fib_per_lst, take_profit_list, take_profit_total

cases = [
    # long
    (
        # OrderProfitFormSimple
        OrderProfitFormSimple(0.1, 1000, 0, 1),
        # TakeProfit list
        [
            TakeProfit(volume=150.0, price=1.382, percentage=0.15),
            TakeProfit(volume=300.0, price=1.618, percentage=0.30),
            TakeProfit(volume=400.0, price=2, percentage=0.40),
            TakeProfit(volume=150.0, price=2.618, percentage=0.15),
            TakeProfit(volume=0, price=3, percentage=0),
        ],
        # total profit
        17854.0,
    ),
    # short
    (
        OrderProfitFormSimple(1, 1000, 1, 0),
        [
            TakeProfit(volume=150.0, price=-0.382, percentage=0.15),
            TakeProfit(volume=300.0, price=-0.618, percentage=0.30),
            TakeProfit(volume=400.0, price=-1, percentage=0.40),
            TakeProfit(volume=150.0, price=-1.618, percentage=0.15),
            TakeProfit(volume=0, price=-2, percentage=0),
        ],
        # total profit
        1885.4,
    ),
    # long normal
    (
        OrderProfitFormSimple(1.784, 665, 1.67, 1.89),
        [
            TakeProfit(volume=99.75, price=1.974, percentage=0.15),
            TakeProfit(volume=199.5, price=2.026, percentage=0.3),
            TakeProfit(volume=266.0, price=2.11, percentage=0.4),
            TakeProfit(volume=99.75, price=2.246, percentage=0.15),
            TakeProfit(volume=0, price=2.33, percentage=0),
        ],
        # total profit
        112.13,
    ),
    # short normal
    (
        OrderProfitFormSimple(1.89, 665, 1.931, 1.784),
        [
            TakeProfit(volume=99.75, price=1.728, percentage=0.15),
            TakeProfit(volume=199.5, price=1.693, percentage=0.3),
            TakeProfit(volume=266.0, price=1.637, percentage=0.4),
            TakeProfit(volume=99.75, price=1.546, percentage=0.15),
            TakeProfit(volume=0, price=1.49, percentage=0),
        ],
        # total profit
        83.11,
    ),
    (
        OrderProfitFormSimple(2636.0, 411.88, 2742.0, 2646.0),
        [
            TakeProfit(volume=61.782, price=2609.328, percentage=0.15),
            TakeProfit(volume=123.564, price=2586.672, percentage=0.3),
            TakeProfit(volume=164.752, price=2550.0, percentage=0.4),
            TakeProfit(volume=61.782, price=2490.672, percentage=0.15),
            TakeProfit(volume=0.0, price=2454.0, percentage=0),
        ],
        # total profit
        11.72,
    ),
]


@pytest.mark.parametrize(("order_profit", "take_profits"), [(case[0], case[1]) for case in cases])
def test_take_profit_list(order_profit, take_profits):
    tp_fib = take_profit_list(order=order_profit)
    assert tp_fib == take_profits


@pytest.mark.parametrize(("order_profit", "take_profits", "total_profit"), cases)
def test_take_profit_total(order_profit, take_profits, total_profit):
    total = take_profit_total(order=order_profit, take_profits=take_profits)
    assert total == total_profit


@pytest.mark.parametrize(
    ("tpp", "expect"),
    [
        (
            ProfitPossibility.LOW,
            [
                TakeProfitFibPer(fib_mark=1.382, vol_per=0.15),
                TakeProfitFibPer(fib_mark=1.618, vol_per=0.4),
                TakeProfitFibPer(fib_mark=2, vol_per=0.4),
                TakeProfitFibPer(fib_mark=2.309, vol_per=0.05),
                TakeProfitFibPer(fib_mark=3, vol_per=0),
            ],
        ),
        (
            ProfitPossibility.MARKET,
            [
                TakeProfitFibPer(fib_mark=1.382, vol_per=0.15),
                TakeProfitFibPer(fib_mark=1.618, vol_per=0.3),
                TakeProfitFibPer(fib_mark=2, vol_per=0.4),
                TakeProfitFibPer(fib_mark=2.618, vol_per=0.15),
                TakeProfitFibPer(fib_mark=3, vol_per=0),
            ],
        ),
        (
            ProfitPossibility.AGGRESSIVE,
            [
                TakeProfitFibPer(fib_mark=1.382, vol_per=0.1),
                TakeProfitFibPer(fib_mark=1.618, vol_per=0.2),
                TakeProfitFibPer(fib_mark=2, vol_per=0.45),
                TakeProfitFibPer(fib_mark=2.618, vol_per=0.15),
                TakeProfitFibPer(fib_mark=3, vol_per=0.1),
            ],
        ),
    ],
)
def test_get_profit_percentage_lst(tpp: ProfitPossibility, expect: list[float]):
    assert get_take_profit_fib_per_lst(tpp) == expect



import pytest

from tradex.constants.listting import TP_FIB, TP_PER_AGGRESSIVE, TP_PER_LOW, TP_PER_MARKET


@pytest.mark.parametrize(
    ("tp_fib", "tp_per"),
    [
        (TP_FIB, TP_PER_MARKET),
        (TP_FIB, TP_PER_LOW),
        (TP_FIB, TP_PER_AGGRESSIVE),
    ],
)
def test_profit_percentage_same_size(tp_fib: list[float], tp_per: list[float]):
    assert len(tp_fib) == len(tp_per)


@pytest.mark.parametrize(
    "tp_fib",
    [
        TP_PER_MARKET,
        TP_PER_LOW,
        TP_PER_AGGRESSIVE,
    ],
)
def test_tp_profit_100_percentage(tp_fib: list[float]):
    assert sum(tp_fib) == 1.0, f"{tp_fib} must equal to one"



from dataclasses import dataclass


@dataclass
class Margin:
    """Margin of a future trade, using the size and leverage values

    :param size: total size of amount to be traded, will be the :class:`Position.position` value after we
        order the trade
    :param leverage: leverage of the trade
    """

    size: float
    leverage: int

    @property
    def margin(self) -> float:
        return self.size / self.leverage

    def __post_init__(self):
        if self.size < 0:
            raise ValueError("Size must be positive")
        if self.leverage < 0:
            raise ValueError("Leverage must be positive")

    def __repr__(self):
        return f"{self.__class__.__name__}(size={self.size!r}, leverage={self.leverage!r}, margin={self.margin!r})"



from dataclasses import dataclass

from tradex.constants.enuming import Direction


@dataclass
class OrderProfitFormSimple:
    price: float
    size: float

    # fib mark to calculate fibonacci retracement
    zero_mark: float
    middle_mark: float

    def __post_init__(self):
        if self.price <= 0:
            raise ValueError("Price must grater than 0")

    @property
    def order_type(self) -> Direction:
        return Direction.LONG if self.middle_mark > self.zero_mark else Direction.SHORT

    @property
    def order_price(self) -> float:
        return self.price * self.size



from dataclasses import dataclass

from tradex.constants.enuming import Direction


@dataclass
class Size:
    """Size of a trade, using the buy, defend and profit values

    :param buy: price at which the asset was bought
    :param defend: price at which the asset was defended, when price at the defend we should sell it
    :param loss: max loss for the trade
    """

    buy: float
    defend: float
    loss: float

    def __post_init__(self):
        if self.buy < 0 or self.defend < 0 or self.loss < 0:
            raise ValueError("Size arguments cannot be negative, check your profit settings")
        self.diff = self.buy - self.defend if self._is_long else self.defend - self.buy
        if self.diff == 0:
            raise ValueError("Price difference cannot be zero, check your buy and defend prices")

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(buy={self.buy!r}, defend={self.defend!r}, loss={self.loss!r}, "
            f"direction={self.direction.value}, size={self.size!r})"
        )

    @property
    def _is_long(self) -> bool:
        return self.buy > self.defend

    @property
    def direction(self) -> Direction:
        return Direction.LONG if self._is_long else Direction.SHORT

    @property
    def size(self) -> float:
        return round(self.loss * (self.buy / self.diff), 2)



import pytest

from tradex.utils.margin import Margin


@pytest.mark.parametrize(
    ("size", "leverage", "margin"),
    [
        (10000, 10, 1000),
        (10000, 1, 10000),
        (3357.4, 125, 26.8592),
    ],
)
def test_margin(size, leverage, margin):
    m = Margin(size=size, leverage=leverage)
    assert margin == m.margin, f"Expected margin is {margin} but got {m.margin}"


@pytest.mark.parametrize(
    ("size", "leverage"),
    [
        (10000, -10),
        (-10000, 10),
    ],
)
def test_margin_invalid(size, leverage):
    with pytest.raises(ValueError, match="must be positive"):
        Margin(size=size, leverage=leverage)



from binance.spot import Spot

client = Spot()

# Get server timestamp
# Get klines of BTCUSDT at 1m interval
# Get last 10 klines of BNBUSDT at 1h interval

# API key/secret are required for user data endpoints
# client = Spot(api_key='<api_key>', api_secret='<api_secret>')
#
# # Get account and balance information
# print(client.account())
#
# # Post a new order
# params = {
#     'symbol': 'BTCUSDT',
#     'side': 'SELL',
#     'type': 'LIMIT',
#     'timeInForce': 'GTC',
#     'quantity': 0.002,
#     'price': 9500
# }
#
# response = client.new_order(**params)
# print(response)



from dataclasses import dataclass


@dataclass
class TPSL:
    volume: float
    price: float
    percentage: float


@dataclass
class TakeProfit(TPSL):
    pass


@dataclass
class StopLoss(TPSL):
    pass


@dataclass
class TakeProfitFibPer:
    fib_mark: float
    vol_per: float

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.fib_mark}, {self.vol_per})"



from tradex.constants.listting import FIB


def fib_retrace_standalone(*, zero: float, middle: float, fib_val: float) -> float:
    return round(zero + (middle - zero) * fib_val, 3)


def fib_retrace(*, zero: float, middle: float, fib_stand: list[float]) -> list[float]:
    """Fibonacci retracement levels for a given range."""
    return [fib_retrace_standalone(zero=zero, middle=middle, fib_val=level) for level in fib_stand]


def fib_retrace_default(*, start: float, middle: float) -> list[float]:
    """Fibonacci retracement levels for default range."""
    return fib_retrace(zero=start, middle=middle, fib_stand=FIB)



from tradex.constants.enuming import Direction, ProfitPossibility
from tradex.constants.listting import TP_FIB, TP_FIB_LOW, TP_PER_AGGRESSIVE, TP_PER_LOW, TP_PER_MARKET
from tradex.models.order import OrderProfitFormSimple
from tradex.models.take_profit import TakeProfit, TakeProfitFibPer
from tradex.utils.fibonacci import fib_retrace_standalone


def get_take_profit_fib_per_lst(tpp: ProfitPossibility) -> list[TakeProfitFibPer]:
    possibility = None

    # take profit fibonacci levels are same in Market and Aggressive, but Low is different
    take_profit_fib = TP_FIB
    if tpp == ProfitPossibility.MARKET:
        possibility = TP_PER_MARKET
    elif tpp == ProfitPossibility.LOW:
        possibility = TP_PER_LOW
        take_profit_fib = TP_FIB_LOW
    elif tpp == ProfitPossibility.AGGRESSIVE:
        possibility = TP_PER_AGGRESSIVE
    else:
        raise ValueError("Parameter %s is unacceptable", tpp)
    return [TakeProfitFibPer(fib_mark=fib, vol_per=per) for fib, per in zip(take_profit_fib, possibility, strict=False)]


def take_profit_list(
    *, order: OrderProfitFormSimple, tpp: ProfitPossibility | None = ProfitPossibility.MARKET
) -> list[TakeProfit]:
    take_profit_fib_per = get_take_profit_fib_per_lst(tpp)

    return [
        TakeProfit(
            volume=round(order.size * tpfp.vol_per, 3),
            price=fib_retrace_standalone(zero=order.zero_mark, middle=order.middle_mark, fib_val=tpfp.fib_mark),
            percentage=tpfp.vol_per,
        )
        for tpfp in take_profit_fib_per
    ]


def take_profit_total(*, order: OrderProfitFormSimple, take_profits: list[TakeProfit]) -> float:
    product_size = order.size / order.price
    sell_total_size = sum([product_size * tp.percentage * tp.price for tp in take_profits])
    if order.order_type == Direction.LONG:
        return round(sell_total_size - order.size, 2)
    else:  # noqa
        return round(order.size - sell_total_size, 2)


# def stop_loss_cal(order: OrderProfitFormSimple,
#                   stop_loss: list[StopLoss]) -> float:
#     total_price = sum([sl.volume * sl.price for sl in stop_loss])
#     return round(order.order_price - total_price, 2)



class DerivativeConverter:
    """Main index price convert into financial derivative price"""

    _default_precision = 2

    def __init__(self, index_price: float, derivative_price: float) -> None:
        self.index_price = index_price
        self.derivative_price = derivative_price

    @property
    def derivative_price_precision(self) -> int:
        return len(str(self.derivative_price).split(".")[1]) if "." in str(self.derivative_price) else self._default_precision

    def tgt_derivative_price(self, tgt_index_price: float) -> float:
        return round(self.derivative_price * tgt_index_price / self.index_price, self.derivative_price_precision)



import pytest

from tradex.constants.enuming import Direction
from tradex.utils.size import Size


@pytest.mark.parametrize(
    ("buy", "defend", "loss", "size", "direction"),
    [
        # long order
        (10000, 9500, 1000, 20000, Direction.LONG),
        (3350, 3122, 1000, 14692.98, Direction.LONG),
        (3350, 10, 1000, 1002.99, Direction.LONG),
        # short order
        (10000, 10500, 1000, 20000, Direction.SHORT),
        (3350, 3500, 1000, 22333.33, Direction.SHORT),
        (10, 3350, 1000, 2.99, Direction.SHORT),
    ],
)
def test_size_prop(buy, defend, loss, size, direction):
    s = Size(buy=buy, defend=defend, loss=loss)
    assert size == s.size, f"Expected size is {size} but got {s.size}"
    assert direction == s.direction, f"Expected direction is {direction} but got {s.direction}"


@pytest.mark.parametrize(
    ("buy", "defend", "loss"),
    [
        (10000, 9500, -1000),
        (10000, -9500, 1000),
        (-10000, 9500, 1000),
    ],
)
def test_size_invalid(buy, defend, loss):
    with pytest.raises(ValueError, match="Size arguments cannot be negative"):
        Size(buy=buy, defend=defend, loss=loss)



from enum import Enum


class Direction(Enum):
    LONG = "long"
    SHORT = "short"


class ProfitPossibility(Enum):
    """Profit possibility for take profit, each possibility has its own fibonacci retracement number.

    The detail fibonacci number related in :mod:`tradex.constants.listting`"""

    LOW = "low"
    MARKET = "market"
    AGGRESSIVE = "aggressive"



import pytest

from tradex.utils.derivative import DerivativeConverter


@pytest.mark.parametrize(
    ("index_price", "derivative_price", "tgt_index_price", "expected"),
    [
        (683.397, 0.78, 1021.3213, 1.17),
        (683.397, 0.782, 1135, 1.299),
        (683.397, 1, 1700, 2.49),
    ],
)
def test_tgt_derivative_price(index_price, derivative_price, tgt_index_price, expected):
    converter = DerivativeConverter(index_price=index_price, derivative_price=derivative_price)
    tgt_derivative_price = converter.tgt_derivative_price(tgt_index_price)
    assert tgt_derivative_price == expected
